require 'web'

require 'web/htmltools/tree'
module HTMLTree # :nodoc: all
  module TreeElement
    def get_elements aname, elements=[]
	    children.each { |element|
        if element.tag == aname
          elements.push element
        else
          element.get_elements aname, elements
        end
	    }
	    elements
    end
  end
end

require 'web/assertinclude'
require 'web/strscanparser'
require 'web/tagparser'
require 'web/formreader.rb'

require 'test/unit/assertions'

module Web
  def Web::setup( options = {} )
    options[:out] ||= ''
    Web::config[:error_style] = :testing
    Web.connection = Web::Connection.new(options)
  end

  def Web::teardown
    Web::config[:error_style] = :default
  end
  
  class Connection
    include Test::Unit::Assertions
    
    # for testing: raises Test::Unit exception if content
    # is not set to the provided string.
    def assert_content expected, message=""
      assert_equal( expected, get_content, message );
    end
    
    def get_formreader #:nodoc:
      return @form_fields_cache ||= FormReader.new( get_content )
    end
    
    def get_form( name ) #:nodoc:
      get_formreader[name]
    end
    
    def get_form_fields(name) #:nodoc:
      get_formreader.get_fields(name)
    end
    
    def get_form_value formname, name #:nodoc:
      form = get_form_fields(formname)
      if form == nil
        raise Web::Error.new("Form '#{formname}' does not exist") 
      end
      
      value = form[name]
      value
    end
    
    # assert output content contains a form
    # that includes the given hash of values.  See Web::Testing
    def assert_form_includes formname, vars
      assert_includes vars, get_form_fields(formname)
    end

    # Raises Test::Unit::AssertionFailedError if the header
    # has not been set to the provided value(s)
    def assert_header( name, values, message="" )
      assert_equal( [values].flatten, get_header(name), message )
    end

    # Throws Test::Unit::AssertionFailedException if
    # cookie values are not present.
    def assert_cookie( name, values, message="" )
      assert_equal( [ values ].flatten, [ cookies_sent[name] ].flatten, message )
    end

        # Check that a particluar template was used 
    def assert_template_used expected, message =""
      _wrap_assertion {
        flunk("No template used") if @templates == nil
        unless @templates.include? expected
          msg = "expected template:<#{expected}> but not one of:<#{@templates.inspect}>"
          flunk(msg)
        end
      }
    end
    
    # Check that a particluar template was not used 
    def assert_template_not_used expected, message =""
      _wrap_assertion {
        if (@templates.kind_of?(Array) && @templates.include?(expected) )
          msg = "template:<#{expected}> not supposed to be used in:<#{@templates.inspect}>"
          flunk(msg)
        end
      }
    end
    
    # Check that these values were included in the output
    def assert_local_includes expected
      assert_includes expected, @local
    end

    def assert_vars_equals
      raise Exception.new("Not yet implemented")
    end

    
  end
  
  # == Purpose
  # The testing module facilitates the testing of Web applications
  # without the overhead of a web server to run.
  #
  # Given these files:
  #
  #    script.rb
  #    test.rb
  #
  # where script.rb is:
  #
  #    #!/usr/bin/ruby
  #    require 'web'
  #
  #    Web::process { 
  #      Web.write "param is #{Web["param"]}"
  #    }
  #
  # and test.rb is:
  #    
  #    require 'web'
  #    require 'test/unit'
  #
  #    class MyAppTest < Test::Unit::TestCase
  #        include Web::Testing
  #
  #        def test_prints_content
  #            do_request "script.rb", "param" => "THIS!"
  #            assert_content "param is THIS!"
  #        end
  #    end
  #
  # Do this to run tests:
  #
  #     ruby test.rb
  #
  # If you have a more complicated app, where the tests live in
  # a different place than your scripts, you can use:
  #
  #    Web::docroot = path
  #
  # To tell narf where to find your cgi scripts.
  #
  #
  # === Testing with Templates
  # 
  # Using Narflates you can test functionality without having to
  # do lengthly string comparisons.  For example, create the following
  # file in 'mytemplate.html'
  #
  #     <html>
  #       <body>
  #         {$myvar}
  #       </body>
  #     </html>
  # 
  # Create a 'script.rb' as follows:
  #
  #     #!/usr/bin/narf
  #     require 'web'
  #
  #     Web::process{
  #       Web.print_template "mytemplate.html", { "myvar" => "Hello World" }
  #     }
  #
  # Now, we can check that the right values got displayed without
  # needing to check that the template is correct as a side effect.
  # Save this into 'test.rb' and run it:
  #
  #     require 'web'
  #
  #     class MyAppTest < Test::Unit::TestCase
  #         include Web::Testing          # adds the modules
  #
  #         def test_prints_content
  #             do_request "script.rb"
  #             assert_vars_includes "myvar" => "Hello World"
  # 	      end
  #     end
  #
  # === Testing Forms
  #
  # The following example demonstrates testing a simple HTML form.
  # Creating mytemplate.html as:
  #
  #
  #     <html>
  #       <body>
  #         <form name="myform">
  #           <input type="text" name="foo">
  #           <input type="submit" name="submit" value="Submit">
  #       </body>
  #     </html>
  #
  # To print this form and handle a submit save this as 'script.rb':
  #
  #     #!/usr/bin/narf
  #     
  #     require 'web'
  #     
  #     Web::process {
  #       if Web["submit"]   # check to see whether a form was
  #           Web.puts "Form Submitted with value '#{Web["foo"]}'"  
  #       else
  #           Web.print_template "mytemplate.html"
  #       end
  #     }
  #
  # Use this 'test.rb' to test it:
  #
  #     class MyAppTest < Test::Unit::TestCase
  #         include Web::Testing          # adds the modules
  #
  #         def test_prints_content
  # 	          do_request "script.rb"
  #             do_submit "myform", "foo" => "bar"
  #             assert_content "Form Submitted with value '#{Web["foo"]}'"
  # 	      end
  #     end
  #
  #
  # === Test <input type="text|password|hidden"> and <textarea>
  #
  # html:
  #
  #     <form name='aForm'>
  #     <input name="bare">
  #     <input name="named"   type="text"     value="foo">
  #     <input name="pass"    type="password" value="secret">
  #     <input name="obscure" type="hidden"   value="discrete">
  #     <textarea name="big_text">
  #     big paragraph here
  #     </textarea>
  #     </form>
  #
  # assert:
  #
  #     assert_form_includes( 'aForm', "bare"     => "",
  #                                    "named"    => "foo",
  #                                    "pass"     => "secret",
  #                                    "obscure"  => "discrete",
  #                                    "big_text" => "big paragraph here" )
  #
  # submit:
  #
  #     do_submit( 'aForm', "bare"     => "empty",
  #                         "named"    => "bare",
  #                         "pass"     => "shhhhh",
  #                         "obscure"  => "secretive",
  #                         "big_text" => "windbag" )
  #
  # === Test <input type="file">
  #
  # html:
  #
  #     <form name='aForm' enctype='multipart/form-data'>
  #     <input name="upload" type="file">
  #     </form>
  #
  # assert:
  #
  #     assert_form_includes( 'aForm', "upload" => "" )
  #
  # submit:
  #
  #     do_submit( 'aForm', "upload" => Web::Upload.new(
  #                                       File.new( "testfile" ),
  #                                       "content-type",
  #                                       "original-filename" ) )
  #
  # === Test <select> and <input type="radio">
  #
  # 
  #
  # === Test <select multiple> and <input type="checkbox">
  #
  #
  # === Bugs: Unsupported behaviour
  #
  #  The following situations will have unknown results:
  #
  #  * Combining different types of elements into one field;
  #    i.e. <input name="field" type="text"> and <select name="field">
  #  * Comparing <input name="page.name"> and <input name="page.content">
  #    with assert_vars_include("page" => { "name" => ..., "content" => ... } )
  #
  module Testing
    class FormNotFoundException < Exception #:nodoc:
    end

    class FieldNotFoundException < Exception #:nodoc:
    end

    class MustSetDocrootForAbsolutePathException < Exception #:nodoc:
    end

    def write_file( path, content )
      filename = File.join( Web::docroot, path )
      filedir  = File.dirname(filename)
      unless File.exists? filedir
        require 'fileutils'
        FileUtils::mkdir_p(filedir, :verbose => false)
      end
      
      File.open( filename, "w" ) do |f|
        f.write content
      end
    end

    def select( *args )
      hash = {}
      
      if (args.length == 1 and args[0].kind_of? Hash)
        hash = args[0]
      else
        args.each { |e|
          hash[e] = true
        }
      end
      
      hash
    end
    
    @@test_session = nil

    # Reset the session used by the test framework. Call prior to all tests
    # that rely on the session being clean
    def reset_session
      @@test_session = nil
    end
    
    # Run a request, parameters are the name value pairs that would be
    # passed in the query string. The webpath is a document root relative 
    # path to a ruby script.
    def do_request(webpath, parameters={}) 
      options = {}
      @@test_session ||= {}
      options[:session] = @@test_session
      options[:env] = parameters[:env]
      options[:request] = parameters
      options[:document_root] = Web::docroot
      options[:out] = StringIO.new
      script_path, script_name, path_info = get_script_part(webpath)
      
      options[:path_info] = path_info
      options[:script_name] = script_name
      load_request( options, script_path, webpath )
    end
    
    def load_request( options, script_path, webpath ) #:nodoc:
      options[:webpath] = webpath
      ### re-throw errors
      Web::config[:error_style] = :custom
      Web::config[:error_handler] = lambda do |error|
        raise error
      end
      
      # out with the old....
      #Web::connection.close
      Web::connection = nil
      error = nil
      Web::open( options ) do
        Web::load( script_path )
      end
      
      if Web.status == "302"
	      Web.location =~ /(.*)\?(.*)/
        target = $1
        params = Connection::parse_query_string($2)
        
        unless (target =~ /^\//)
          webpath =~ /(.*)#{File.basename(script_path)}/
          target = $1 + target
	      end
        do_request( target, params )
      end
    end
    
    # Submit the form 'formname' with the formfields described in newvalues
    def do_submit( formname, newvalues={} )
      form = Web.get_form(formname)
      
      if form == nil
        #if Web.get_form_fields[formname] == nil
	raise FormNotFoundException.new("Form '#{formname}' does not exist") 
      end
      
      #MultiHashTree::flatten(newvalues).keys.each do |key|
	#unless( Web.get_form_fields[formname].valid_key?(key) )
	 # raise FieldNotFoundException.new( "#{ key } is not present in form" )
	#end
      #end

      webpath = nil
      
      #Web.get_html_tree.get_elements("form").each { |form|
      #  if (form.attribute("name") == formname)
      #    webpath = form.attribute("action")
      #  end
      #}
      webpath = form.action

      unless (webpath)
        webpath = Web.options[:webpath]
      end

      script_path, script_name, path_info = get_script_part(webpath)

      old_connection = Web.connection

      options = {}
      @@test_session ||= {}
      options[:session] = @@test_session
      options[:out] = StringIO.new
      ### patrick
      options[:request] = form.merge_fields( newvalues )
      #options[:request] = Web.get_form_fields[formname].params(MultiHashTree::flatten(newvalues))
      options[:document_root] = old_connection.document_root

      options[:path_info] = path_info     #old_connection.path_info
      options[:script_name] = script_name #old_connection.script_name

      load_request( options, script_path, webpath )
    end

    # Assert that a give template was displayed
    def assert_template_used filename, msg=""
      Web.assert_template_used filename, msg
    end

    # Assert that a give template was not displayed
    def assert_template_not_used filename, msg=""
      Web.assert_template_not_used filename, msg
    end

    # Assert that the values passed in to expected were set on the template
    def assert_vars_includes expected
      Web.assert_vars_includes expected
    end
    
    # Assert that the form displayed contains particular values
    def assert_form_includes formname, expected
      Web::assert_form_includes formname, expected
    end
    
    # Assert that the entire content displayed is equal to expected
    def assert_content expected, msg=""
      Web.assert_content expected, msg
    end
    
    # Assert that the header key has the value 'value'
    def assert_header key, value, msg=""
      Web.assert_header key, value, msg
    end

    # Assert that the cookie key, has the cookie value
    def assert_cookie key, value, msg=""
      Web.assert_cookie key, value, msg
    end
    
    # Assert that a form field has exactly the given options
    # can't assert order, though
    def assert_options( formname, expected={})
      options = Web.get_formreader.get_options(formname)
      expected.each{ |k,v|
        assert_equal( v.sort, options[k].sort )
      }
    end

    def remove_trailing_slashes(filename)  #:nodoc:
      /(.*?)\/?$/.match(filename)[1]
    end

    def get_script_part(webpath)  #:nodoc:
      # two cases:
      #   absolute webpath (requires Web::docroot to be correct)
      #   relative webpath (doesn't require Web::docroot to be correct)
      if ( webpath =~ /^\// )
        prefix = Web::docroot
        raise MustSetDocrootForAbsolutePathException unless prefix
        path = "/"
      else
        prefix = ""
        path = ""
      end

      new_script_path = ""
      webpath.split("/").each do |file|
	file =~ /^([a-zA-Z]{1,1}:)$/

        if (path.empty?)
          newpath = file
        else
          newpath = File.join( path , file ).gsub("//","/") 
        end

        new_script_path = ""
        if (prefix.empty?)
          new_script_path = newpath
        else
          new_script_path = File.join( prefix, newpath ).gsub("//","/") 
        end
        
        if (File.file?(new_script_path))
	  return new_script_path, "/" + File.basename( new_script_path ), webpath[newpath.length...webpath.length]
	end 
	
	path = newpath
      end

      if (!File.exists?(new_script_path) && prefix.empty?)
        return get_script_part( "/" + webpath )
      end

      return new_script_path, "/" + File.basename( new_script_path ), ""
    end

    class SelectHash < Hash #:nodoc:
      def == other
        if (other.kind_of? Hash)
          super(other)
        elsif (other.class == TrueClass)
          result = true
          self.each{ |k, v|
            result = false unless v
          }
          result
        elsif (other.class == FalseClass)
          result = true
          self.each{ |k,v|
            result = false if v
          }
          result
        else
          ### this is wrong
          other.to_s == self.default.join(",")
        end
      end

      def field_value
        values = [ ]
        self.each{ |k,v|
          values.push k if v
        }
        values.join(",")
      end

    end

    # could I get an explanation of this?
    # just for me? please?
    # not to long... probably not longer than this comment here.
    class MultiHashTree # :nodoc:
      attr_reader :fields, :unmodified_fields
      
      def initialize fields = {}
        @fields = {}
        @unmodified_fields = {}
        fields.each do |k,value_array|
          value_array.each{ |v|
            push_field k, v
          }
        end
      end
      
      def valid_key? (aKey)
        unmodified_fields.has_key?(aKey) || fields.has_key?(aKey)
      end
      
      def push_impl hash, name, value
        if name =~ Web::Connection::MULTIPLE_KEY && !value.kind_of?(Web::Testing::SelectHash)
          hash[name] ||= []
          hash[name].push  value
        elsif m = /^(\w+(\[\])?)\.(.+)/.match(name)
          hash[m[1]] ||= {}
          push_impl(hash[m[1]], m[3], value)
        elsif m = /^(\w+)\[(\d+)\].(.+)/.match(name)
          hash[m[1]] ||= []
          hash[m[1]][m[2].to_i] ||= {}
          push_impl hash[m[1]][m[2].to_i], m[3], value
        else
          hash[name] ||= []
          hash[name].push(value)
        end
      end

      def push_field name, value
        @unmodified_fields[name] ||= []
        @unmodified_fields[name].push value
        push_impl @fields, name, value
      end
      
      def params values
        retval = @unmodified_fields.clone
        values.each do |key,value|
          if value.kind_of? SelectHash
            selected_options = []
            value.each{ |k,v|
              selected_options.push k if v
            }
            
            if m = Web::Connection::MULTIPLE_KEY.match(key)
              retval[key] ||= []
              retval[key] = (retval[key] + selected_options).uniq
            else
              retval[key] = selected_options.join(",")
            end
            
          else
            if m = Web::Connection::MULTIPLE_KEY.match(key)
              retval[key] ||= []
              retval[key] << value
            else
              retval[key] = value
            end
          end
        end
        retval
      end
      
      def MultiHashTree.flatten node, newhash = {} , nameroot = "" # :nodoc:
        if node.kind_of? Hash
          node.each do |k,v|
            if nameroot == ""
              varname = k
            else
              varname = "#{nameroot}.#{k}" 
            end
            flatten( v, newhash, varname )
          end
        elsif node.kind_of? Array
          unless node.find { |i| !i.kind_of? String }
            newhash[nameroot] = node
          else
            node.each_with_index do |v,i|
              flatten( v, newhash, "#{nameroot}[#{i}]" )
            end
          end
        else
          newhash[ nameroot ]=node
        end
        newhash
      end
      
    end

  end
end
