require 'strscan'

def resolve array, name, anAlias = ""
  Web::Scope.new(array).resolve name, anAlias
end

class TestConvertArrayToSymbol < Test::Unit::TestCase
  def test_simple_hash
    hash = { "name" => "value" }

    var = resolve( hash, "name" )
    assert_equal "name", var.canonical_name
    assert_equal "value", var.value
  end

  def test_simple_array
    array = ["patrick", "tom"]
    var = resolve( array, "[0]" )
    assert_equal "[0]", var.canonical_name
    assert_equal "patrick", var.value
  end


  class Trivial
    attr_accessor :name
    def initialize( name )
      @name = name
    end

    def == (other)
      (other.kind_of? Trivial) && (self.name == other.name)
    end
  end

  def test_simple_object
    object = Trivial.new( "value" )

    var = resolve( object, "name" )

    assert_equal "name", var.canonical_name
    assert_equal "value", var.value
  end


  def test_nested_hash
    array = { "name" => { "name2" , "value" } }

    var = resolve( array, "name.name2" )
    assert_equal "name.name2", var.canonical_name
    assert_equal "value", var.value
  end

  def test_nested_object
    object = Trivial.new( Trivial.new( "nested" ) )

    var = resolve( object, "name.name" )
    assert_equal "name.name", var.canonical_name
    assert_equal "nested", var.value
  end

  def test_missing_method_on_object
    object = Trivial.new( "value" )

    assert_raises( Web::Scope::VarNotFoundException ) {
      var = resolve( object, "missing_method" )
    }
  end

  def test_missing_method_falls_back_to_parent_scope
    array = { 
      "name"   => [Trivial.new("nested")],
      "hash"   => {"name" => "hash_value"},
      "object" => Trivial.new( "object_value" ) }

    var = resolve( array, "name" )

    looped = false
    var.each do |i|
      var = i.resolve("name")
      assert_equal "name[0].name", var.canonical_name
      assert_equal "nested", var.value		

      var = i.resolve( "hash.name" )
      assert_equal "hash.name", var.canonical_name
      assert_equal "hash_value", var.value

      var = i.resolve("object.name")
      assert_equal "object.name", var.canonical_name
      assert_equal "object_value", var.value		

      assert_raises( Web::Scope::VarNotFoundException ) {
	i.resolve( "missing_method" )
      }

      looped = true
    end
    assert looped
  end


  def test_nested_nested_hash
    array = { "name" => { "name2" => { "name3" , "value" } } }

    var = resolve( array, "name.name2.name3" )
    assert_equal "name.name2.name3", var.canonical_name
    assert_equal "value", var.value
  end

  def test_nested_array
    array = { "name" => [{ "name2" , "value" }]}

    var = resolve( array, "name[0].name2" )
    assert_equal "name[0].name2", var.canonical_name
    assert_equal "value", var.value		
  end

  def test_top_level_array
    array = [{ "name2" , "value" }]

    var = resolve( array, "[0].name2" )
    assert_equal "[0].name2", var.canonical_name
    assert_equal "value", var.value		
  end

  def test_each
    array = { "name" => [{ "name2" , "value" }]}

    var = resolve( array, "name" ,"item")

    looped = false
    var.each do |i|
      var = i.resolve("item.name2")
      assert_equal "name[0].name2", var.canonical_name
      assert_equal "value", var.value		
      looped = true
    end
    assert looped
  end


  def test_nested_each
    array = { "name" => [{ "name2" => [ { "name3" =>"value" } ] } ] }

    var = resolve( array, "name" ,"item")

    looped = false
    loopedj = false
    var.each do |i|
      jvar = i.resolve("item.name2","itemj")
      assert_equal "name[0].name2", jvar.canonical_name

      jvar.each do |j|
	var = j.resolve("itemj.name3")
	assert_equal "name[0].name2[0].name3", var.canonical_name
	assert_equal "value", var.value		
	loopedj = true
      end

      looped = true
    end
    assert looped
    assert loopedj
  end


  def test_nested_each_from_same_scope
    array = { "name" => [{ "name1" => "value" }],
      "name2" => [{ "name3" => "value" }] }

    var = resolve( array, "name" ,"item")

    looped = false
    loopedj = false
    var.each do |i|
      jvar = i.resolve("name2","itemj")
      assert_equal "name2", jvar.canonical_name

      varx = i.resolve("item.name1")
      assert_equal "name[0].name1", varx.canonical_name
      assert_equal "value", varx.value		


      jvar.each do |j|
	var = j.resolve("itemj.name3")
	assert_equal "name2[0].name3", var.canonical_name
	assert_equal "value", var.value		

	var = j.resolve("item.name1")
	assert_equal "name[0].name1", var.canonical_name
	assert_equal "value", var.value		
	loopedj = true
      end

      looped = true
    end
    assert looped
    assert loopedj
  end

  def test_3nested_eachs_from_same_scope
    array = { "name" => [{ "name1" => "value" }],
      "name2" => [{ "name3" => "value" }],
      "name4" => [{ "name5" => "value" }]
    }

    var = resolve( array, "name" ,"item")

    looped = false
    loopedj = false
    loopedk = false
    var.each do |i|
      jvar = i.resolve("name2","itemj")
      assert_equal "name2", jvar.canonical_name

      jvar.each do |j|
	var = j.resolve("itemj.name3")
	assert_equal "name2[0].name3", var.canonical_name
	assert_equal "value", var.value		

	var = j.resolve("item.name1")
	assert_equal "name[0].name1", var.canonical_name
	assert_equal "value", var.value		

	kvar = j.resolve("name4", "itemk")
	assert_equal "name4", kvar.canonical_name
	kvar.each do |k|
	  var = k.resolve( "itemk.name5")
	  assert_equal "name4[0].name5", var.canonical_name
	  assert_equal "value", var.value
	  
	  var = k.resolve("itemj.name3")
	  assert_equal "name2[0].name3", var.canonical_name
	  assert_equal "value", var.value		
	  
	  var = k.resolve("item.name1")
	  assert_equal "name[0].name1", var.canonical_name
	  assert_equal "value", var.value		

	  loopedk = true
	end

	loopedj = true

      end

      looped = true
    end
    assert looped
    assert loopedj
    assert loopedk
  end


  def test_nested_scope
    array = { 
      "name" => 
      [{ "name2" => "value" }],
      "name3" => "value3",
      "name4" => { "name" => "value4" },
      "name5" => [{ "name" => "value5" }] }

    var = resolve( array, "name" )

    looped = false
    var.each do |i|
      var = i.resolve("name2")
      assert_equal "name[0].name2", var.canonical_name
      assert_equal "value", var.value		

      var = i.resolve("name3")
      assert_equal "name3", var.canonical_name
      assert_equal "value3", var.value		

      var = i.resolve("name4.name")
      assert_equal "name4.name", var.canonical_name
      assert_equal "value4", var.value		

      var = i.resolve("name5[0].name")
      assert_equal "name5[0].name", var.canonical_name
      assert_equal "value5", var.value		
      looped = true

    end
    assert looped
  end

  # update
  def test_update
    initial = { "name" => "oldvalue" }

    submitted = { "name" => "newvalue" }

    Web::Scope.new(initial).update( submitted )

    assert_equal( initial, submitted )
  end

  def test_update_object
    initial = Trivial.new("oldvalue")

    submitted = { "name" => "newvalue" }

    Web::Scope.new(initial).update( submitted )

    assert_equal( Trivial.new("newvalue"), initial )
  end

  def test_update_ignores_extra_submissions
    initial = { "name" => "oldvalue" }

    submitted = { "name" => "newvalue", "something" => "else" }

    Web::Scope.new(initial).update( submitted )

    assert_equal( {"name" => "newvalue"}, initial )
  end

  def test_update_nested
    initial = { "people" => { "sally" => "happy", "dave" => "sad" } }
    submitted = { "people.sally" => "angry", "people.fool" => "ignored" }

    Web::Scope.new(initial).update( submitted )

    assert_equal( { "people" => { "sally" => "angry", "dave" => "sad" } },
		 initial )
  end

  def test_update_array
    initial =   ["sally", "dave"]
    submitted = { "[0]" => "tom" }

    Web::Scope.new(initial).update( submitted )
    
    assert_equal( ["tom","dave"], initial )
  end

  class Restricted
    attr_reader :name
    def initialize( name )
      @name = name
    end
    def == (other)
      (other.kind_of? Restricted) && (self.name == other.name)
    end
  end

  def test_fail_invalid_assignment
    initial = Restricted.new( "value" )
    submitted = { "name" => "new value" }

    assert_raises( if (RUBY_VERSION =~ /1\.7/ || RUBY_VERSION =~ /1\.8/ )
		     NoMethodError             # the 1.7-8 exception has a better name
		   else
		     NameError
		   end ) {
      update( initial, submitted )
    }
  end

  # Error reporting
  

  # replace code in parser

  # Then we can evaluate arbitrary expressions in a page, and test what the results were
  # update the the results with values submitted from a form
end



#Two Tables:
#VariableTable
#.resolve("name.value")
#.canonical_name



class TestMultiHashTree2 < Test::Unit::TestCase
  def test_flatten_single
    assert_equal({"a"=>"b"}, Web::Testing::MultiHashTree::flatten({ "a" => "b"}))
  end

  def test_flatten_single
    assert_equal({"a"=>["b"]}, Web::Testing::MultiHashTree::flatten({ "a" => ["b"]}))
  end

  def test_flatten_multiple
    assert_equal( {"a[]"=>["b", "c"]}, Web::Testing::MultiHashTree::flatten({"a[]"=>["b", "c"]}) )
  end

  def test_flatten_hash
    assert_equal( {"top.b" => "c", "top.x" => "y"}, Web::Testing::MultiHashTree::flatten({"top"=>{"b"=>"c","x"=>"y"}}))
  end

  def test_flatten_array
    assert_equal( {  "items[0].a" => "b",
		   "items[1].c" => "d" }, Web::Testing::MultiHashTree::flatten({ "items" => [ 
								{ "a" => "b"} ,
								{ "c" => "d"} ] }
							    ) )
  end
end

class TestMultiHashTree3 < Test::Unit::TestCase
  def test_add_simple
    fs = Web::Testing::MultiHashTree.new
    fs.push_field "name","value"
    fs.push_field "name2","value2"
    
    assert_equal({ "name" => ["value"],
		   "name2" => ["value2"] }, fs.fields)	
  end

  def test_construct_from_hash
    fs = Web::Testing::MultiHashTree.new({ "name" => ["value"] , "name2" => ["value2"] })
    
    assert_equal({ "name" => ["value"],
		   "name2" => ["value2"] }, fs.fields)	
  end

  def test_add_array
    fs = Web::Testing::MultiHashTree.new
    fs.push_field "name[]","value"
    fs.push_field "name[]","value2"
    
    assert_equal({ "name[]" => [ "value", "value2" ] }, fs.fields)	

  end

  def test_add_hashed
    fs = Web::Testing::MultiHashTree.new
    fs.push_field "name.a","value"
    fs.push_field "name.b","value2"
    
    assert_equal({ "name" => { "a" => ["value"], "b" => ["value2"]}}, fs.fields)	
  end

  def test_nested_hash
    fs = Web::Testing::MultiHashTree.new
    fs.push_field "name.a.c","value"
    fs.push_field "name.b","value2"
    
    assert_equal({ "name" => { "a" => { "c" => ["value"]} , "b" => ["value2"]}}, fs.fields)	
  end


  def test_add_array
    fs = Web::Testing::MultiHashTree.new
    fs.push_field "name[0].a","value"
    fs.push_field "name[1].a","value2"
    
    assert_equal({ "name" => [ { "a" => ["value"] } , { "a" => ["value2"] }]}, fs.fields)	
  end

  def test_add_hash_in_array
    fs = Web::Testing::MultiHashTree.new
    fs.push_field "name[0].a.b","value"
    fs.push_field "name[1].a.b","value2"
    
    assert_equal({ "name" => [ { "a" => { "b" => ["value"] } } , { "a" => { "b" => ["value2"] }}]}, fs.fields)	
  end

end
