class TestPHPRBTemplate < Test::Unit::TestCase
  include Web::Testing

  def setup
    teardown

    Web::setup( :document_root => get_test_dir, :out => '' )

    template_dir = File.expand_path((File.join("#{get_test_dir}","templates")))
    
    unless Web::config[:load_path].include? template_dir
      Web::config[:load_path] << template_dir
    end
    
    Dir.mkdir(get_test_dir)
    Dir.mkdir(template_dir)

    @old_default_handler ||= Web::config['load_suffix'].default
    Web::config['load_suffix'].default = lambda do |script|
      PHPRB::load(script)
    end
  end
    
  def teardown
    Web::teardown
    @old_default_handler ||= Web::config['load_suffix'].default
    Web::config['load_suffix'].default = @old_default_handler
    
    if File.exists? get_test_dir
      rm_rf(get_test_dir)
    end
  end


  
  def get_test_dir
    File.expand_path(File.join(temp_dir,"test_dir"))
  end
  
  def temp_file filename
    File.expand_path(File.join(get_test_dir,"templates",filename))
  end

  def write_template contents, filename = "test.html"
    File.open( File.expand_path(File.join(get_test_dir,"templates",filename))  ,"w") { | file | file.write contents }
  end

  def handle_request cgi, filename = "test.html"
    cgi.print_template(  filename, { "var" => cgi["displayVar"] } )
  end

  def write_file file, filename="app.rb"
    File.open( File.expand_path(File.join("#{get_test_dir}",filename)), "w" ) { |f| f.write(file) }
  end

  def test_template_file
    write_template( "here's a test" )
    assert_equal( "#{temp_dir}/test_dir/templates/test.html",
		  Web::template_file( "test.html" ) )
    
    teardown
    setup
      
    begin
      File.open( "test.html", "w" ) { |f|
        f.write("here's a test")
      }
      assert_equal( "./test.html",
                    Web::template_file( "test.html" ) )
    ensure
      File.delete( "test.html" )
    end
    
  end

  def test_template_dir_precedence
    write_template( "here's a test" )
    begin
      File.open( "test.html", "w" ) { |f|
        f.write("here's a test")
      }
      assert_equal( "./test.html",
                    Web::template_file( "test.html" ) )
    ensure
      File.delete( "test.html" )
    end
  end

  # ========= Variable Resolution ==========
  def test_print_vars
    write_template "x<?= Web::local['foo'] ?>y<?= Web::local['bar'] ?>z"
    vars = { "foo" => "FOO", "bar" => "BAR" }
    Web::options[:local] = vars
    Web::load( "test.html" )
    Web::close
               
    Web::assert_content "xFOOyBARz"
  end

  def test_xml_embeding
    write_template "x<?= 'foo' ?>y<? Web::write('bar') ?>z"
    Web::options[:xml] = true
    Web::load( "test.html" )
    Web::close
    Web::assert_content "x<?= 'foo' ?>y<? Web::write('bar') ?>z"
  end

  def test_xml_embeding2
    write_template "x<?= 'foo'\n\n ?>y<? Web::write('bar') ?>z"
    Web::load( "test.html")
    Web::close
    Web::assert_content "xfooybarz"
  end

  def test_xml_open_tags
    write_template "x<?ruby= 'foo' ?>y<?ruby Web::write('bar') ?>z"
    Web::load( "test.html" )
    Web::close
    Web::assert_content "xfooybarz"
  end

  def test_comments
    write_template "x<!--- this will be missing --->y<!--- this too --->z"
    Web::load( "test.html" )
    Web::close
    Web::assert_content "xyz"
  end

  def test_comments_with_internal_tags
    write_template "x<!--- this will be <?= missing ?> --->y<!--- this too --->z"
    Web::load( "test.html" )
    Web::close
    Web::assert_content "xyz"
  end

  def deprecated_test_throw_exception_if_var_doesnt_exist
    write_template "x<?= Web::local['foo'] ?>y<?= Web::local['bar'] ?>z"
    vars = { "foo" => "FOO" }
    
    assert_raises(NameError) {
      Web::options[:local] = vars
      Web::load( "test.html" )
                 
    }
  end

  def test_nil_values_display_nothing
    write_template "x<?= Web::local['foo'] ?>yz"
    vars = { "foo" => nil }
    Web::options[:local] = vars
    Web::load( "test.html" )
    Web::close
    Web::assert_content "xyz"
  end


  class TestClass
    attr_reader :var
    
    def initialize var
	    @var = var
    end
    
    def foo
	    TestClass.new("XXX")
    end
  end
  
  def test_print_method
    write_template "x<?= Web::local['foo'].var ?>y<?= Web::local['bar'] ?>z"
    vars = { "foo" => TestClass.new("FOO"), "bar" => "BAR" }
    
    Web::options[:local] = vars
    Web::load( "test.html" )
    Web::close
    Web::assert_content "xFOOyBARz"
  end

  def test_template
    write_template "<html><body><h1>hi</h1></body></html>"
    
    Web::load( "test.html" )
    Web::close
    Web::assert_template_used( "test.html" )
  end

  def test_exception_thrown_if_no_template_used
    write_template "<html><body><h1>{$var}</h1></body></html>"

    assert_raises(Test::Unit::AssertionFailedError) do
      Web::assert_template_used( "/test.html" )
    end
  end

  def test_template_fails_assertion_with_wrong_template
    write_template "<html><body><h1>{$var}</h1></body></html>"
  
    Web::load( "test.html" )
    Web::close

    assert_raises( Test::Unit::AssertionFailedError ) do
      Web::assert_template_used( "test_not_it.html" )
    end
      
  end
  
  def test_assert_template_not_used
    write_template "<html><body><h1>foo</h1></body></html>"
    Web::load( "test.html" )
    Web::assert_template_not_used( "foo.html" )
  end
  
  def test_assert_template_not_used_with_no_template_used
    write_template "<html><body><h1>foo</h1></body></html>"
    
    Web::assert_template_not_used( "test.html" )
  end
  
  def test_basic_load
    write_template "blah blah blah"
    Web::load( "test.html" )
    Web::close
    Web::assert_content "blah blah blah"
  end
  
  def test_print_template_survives_enormous_templates
    bad_template =  ""
    
    1.upto(1000){ |i|
      bad_template += "this template sucks\n"
    }
    
    write_template( bad_template )
    Web::load( "test.html" )
    Web::close

    Web::assert_content bad_template	
  end
end
