require 'test/web/mod_ruby_stub'

class TestRequest < Test::Unit::TestCase
  
  def env
    ENV
  end

  def raw_post_data
    $stdin
  end

  def assert_keys( expected, found )
    passes = true
    expected.each { |k, v|
      passes = false unless found[k] == v
    }
    passes
  end
  
  def setup
    @stdin_back = $stdin
  end
  
  def teardown
    ENV["QUERY_STRING"  ] = nil
    ENV["REQUEST_METHOD"] = nil
    ENV["CONTENT_TYPE"  ] = nil
    ENV["CONTENT_LENGTH"] = nil
    ENV["HTTP_COOKIE"   ] = nil
    ENV["COOKIE"        ] = nil
    
    @get = @post = @cookies = nil
    
    $stdin = @stdin_back
    
    if defined? MOD_RUBY
      Object.module_eval{ remove_const :MOD_RUBY }
    end
    @cgi = nil
  end
  
  def params
    { "string" => "aString",
     "array[]"  => ["one", "two", "three"],
     "file"   => Web::Upload.new( File.new( "test/web/testfiles/text" ),
                                  "content-type",
                                  "original-filename" ) }
  end
  
  def cookie
    { "aCookie" => "Peanut Butter" }
  end
  
  def cgi
    @cgi ||= Web::Connection.new( :request => params,
                                  :cookie => cookie,
                                  :out => self.out )
  end
  
  def out
    @out ||= ""
  end
  
  def test_mock_cookies
    assert_includes( Web::Connection.normalize(cookie), cgi.cookie )
  end
  
  def test_convenience_defaults_to_param
    assert_equal( "aString", cgi["string"] )
  end
  
  def test_convenience_returns_array
    assert_equal( ["one", "two", "three"], cgi["array[]"] )
  end
  
  def test_setting_params
    assert_equal( "aString", cgi["string"] )
    assert_equal( ["one", "two", "three"], cgi["array[]"] )
    
    cgi["array[]"] = ["four", "five", "six"]
    cgi["string"]  = "another_string"
    
    assert_equal( "another_string", cgi["string"] )
    assert_equal( ["four", "five", "six"], cgi["array[]"] )

  end
  
  def test_single_param
    assert_equal( "aString", cgi.single_param("string") )
    assert_equal( "one,two,three", cgi.single_param("array[]") )
    assert_equal( "original-filename", cgi.single_param("file").original_filename )
    
    assert_equal( "", cgi.single_param("non-existent") )
  end

  def test_multiple_params
    assert_equal( ["aString"], cgi.request["string"] )
    assert_equal( ["one", "two", "three"], cgi.request["array[]"] )
    assert_equal( "original-filename", cgi.request["file"][0].original_filename )
    assert_equal( 1, cgi.request["file"].length)
    
    assert_equal( [], cgi.request["non-existent"] )
    
    Web.connection = cgi
    
    assert_equal( ["aString"], Web::request["string"] )
    assert_equal( ["one", "two", "three"], Web::request["array[]"] )
    assert_equal( "original-filename", Web::request["file"][0].original_filename )
    assert_equal( 1, Web::request["file"].length)
    
    assert_equal( [], Web::request["non-existent"] )
  end

  def test_request
    assert_equal( ["aString"], cgi.request["string"] )
    assert_equal( ["one", "two", "three"], cgi.request["array[]"] )
    assert_equal( "original-filename", cgi.request["file"][0].original_filename )
    assert_equal( 1, cgi.request["file"].length)
    
    assert_equal( [], cgi.request["non-existent"] )
    
    Web.connection = cgi
    
    assert_equal( ["aString"], Web::request["string"] )
    assert_equal( ["one", "two", "three"], Web::request["array[]"] )
    assert_equal( "original-filename", Web::request["file"][0].original_filename )
    assert_equal( 1, Web::request["file"].length)
    
    assert_equal( [], Web::request["non-existent"] )
  end

  def test_keys
    assert_equal( ["string", "array[]", "file"].sort, cgi.keys.sort )
  end

  def test_has_key?
    assert cgi.key?( "string" )
    assert cgi.key?( "array[]" )
    assert cgi.key?( "file" )
    
    assert((cgi.key? "missing key") == false)
  end

  def test_parse_query_string
    assert_equal({}, {})
    assert_keys( {"query1"=>["~value1"], "query2"=>["value1", "value2"] },
                 Web::Connection.parse_query_string('query1=%7Evalue1&query2=value1&query2=value2') )
    
  end

  def test_parse_nil_cookie
    assert_keys( {},  cgi.parse_cookie(nil) )
    assert_equal( [], cgi.parse_cookie(nil).default )
  end

  def test_parse_cookie
    assert_equal( {"cookie1"=>["~value1"], "cookie2"=>["value1", "value2"]}.inspect,
                  cgi.parse_cookie("cookie1=%7Evalue1; cookie2=value1; cookie2=value2").inspect )
  end

  def test_parse_multiple_cookies
    assert_equal( {"cookie"=>["~value1", "value 2"]}.inspect,
                  cgi.parse_cookie("cookie=%7Evalue1&value%202").inspect )
  end

  def test_parse_blank_cookie
    assert_keys( {"cookie"=>[]},
                 cgi.parse_cookie("cookie=") )
  end

  def test_parse_get_and_post
    ENV["REQUEST_METHOD"] = "POST"
    ENV["QUERY_STRING"]   = "key=get_value_1&get_only=get_value_2"
    ENV["CONTENT_LENGTH"] = "40"
    File.open( "test/web/testfiles/post.and.get.data" ){ |post_data|
      $stdin = post_data

      cgi = Web::Connection.new

      assert_equal( {"key" => ["get_value_1"],  "get_only"  => ["get_value_2"] },
                    cgi.get )
      assert_equal( {"key" => ["post_value_1"], "post_only" => ["post_value_2"]},
                    cgi.post )
      assert_equal( {"key" => ["post_value_1"], "get_only"  => ["get_value_2"], "post_only" => ["post_value_2"]},
                    cgi.request )
    }
  end

  def test_cgi_parses_by_default
    ENV["REQUEST_METHOD"] = "GET"
    ENV["QUERY_STRING"]   = "key=a%20value&anotherKey=another%20value"
    assert_keys( {"key" => ["a value"], "anotherKey" => ["another value"] },
                 Web::Connection.new.get )
    assert_keys( {"key" => ["a value"], "anotherKey" => ["another value"] },
                 Web::Connection.new.request )
  end
  
  def test_cgi_parses_http_cookie
    ENV['HTTP_COOKIE'] = "cookie1=%7Evalue1; cookie2=value1; cookie2=value2"
    assert_keys( {"cookie1"=>["~value1"], "cookie2"=>["value1", "value2"]},
                 Web::Connection.new.cookie )
    assert_keys( {"cookie1"=>["~value1"], "cookie2"=>["value1", "value2"]},
                 Web::Connection.new.request )
  end

  def test_cgi_parses_cookie
    ENV['COOKIE'] = "cookie1=%7Evalue1; cookie2=value1; cookie2=value2"
    
    assert_keys( {"cookie1"=>["~value1"], "cookie2"=>["value1", "value2"]},
                 Web::Connection.new.cookie )
  end

  def test_cgi_only_parses_cookies_once
    ENV['HTTP_COOKIE'] = "cookie1=%7Evalue1; cookie2=value1; cookie2=value2"
    ENV['COOKIE'] = "cookie1=%7Evalue1; cookie2=value1; cookie2=value2"
    assert_keys( {"cookie1"=>["~value1"], "cookie2"=>["value1", "value2"]},
                 Web::Connection.new.cookie )
  end
  
  def test_parse_empty_query_string
    ENV["REQUEST_METHOD"] = "GET"
    cgi = Web::Connection.new
    assert_equal( {}, cgi.request )
  end
  
  def test_parse_get_query_string
    ENV["REQUEST_METHOD"] = "GET"
    ENV["QUERY_STRING"]   = "key=a%20value&anotherKey=another%20value"
    cgi = Web::Connection.new
    assert_equal( {"key" => ["a value"], "anotherKey" => ["another value"]}, cgi.request )
  end

  def test_parse_get_multiples
    ENV["REQUEST_METHOD"] = "GET"
    ENV["QUERY_STRING"]   = "key=a%20value&key=another%20value"
    cgi = Web::Connection.new
    assert_equal( {"key" => ["a value", "another value"]}, cgi.request )
  end

  def test_parse_head_like_get
    ENV["REQUEST_METHOD"] = "HEAD"
    cgi = Web::Connection.new
    assert_equal( {}, cgi.request )
    
    teardown

    ENV["REQUEST_METHOD"] = "HEAD"
    ENV["QUERY_STRING"]   = "key=a%20value&anotherKey=another%20value"
    cgi = Web::Connection.new
    assert_equal( {"key" => ["a value"], "anotherKey" => ["another value"]},
                  cgi.request )
  end

  def test_parse_post
    ENV["REQUEST_METHOD"] = "POST"
    ENV["CONTENT_LENGTH"] = "40"
    File.open( "test/web/testfiles/post.data" ){ |post_data|
      $stdin = post_data
      cgi = Web::Connection.new

      assert_equal( "key=a%20value&anotherKey=another%20value", cgi.raw_post_data.read )
      assert_keys( {"key" => ["a value"], "anotherKey" => ["another value"]},
                   cgi.post )
      assert_keys( {"key" => ["a value"], "anotherKey" => ["another value"]},
                   cgi.request )
    }
  end

  # TODO: add multipart tests for mozilla, netscape 4, explorer 5
  def test_parse_multiparts
    multipart_test(177696,
                   '----------0xKhTmLbOuNdArY',
                   'mac-safari-fallow_dump.bin' )
  end

  def multipart_test( content_length, boundary, datafile )
    ENV["REQUEST_METHOD"] = "POST"
    ENV["CONTENT_LENGTH"] = content_length.to_s
    ENV["CONTENT_TYPE"  ] = "multipart/form-data, boundary=#{ boundary }"
    
    @stdin = $stdin
    
    File.open( "test/web/testfiles/#{ datafile }" ) do |post_data|
      $stdin = post_data
      multipart_asserts
    end
  ensure
    $stdin = @stdin
  end

  def multipart_asserts
    progress_events = [ ]
    Web::config[:multipart_progress_hook] = lambda do |total, read|
      progress_events.push [total, read]
    end
    params = Web::open(:out=>'') do |connection|
      connection.request
    end
    
    image = 'fallow_keep.png.450x450.2005-07-24.jpg'
    assert_equal( ["key", "file"], params.keys )
    assert_equal( "value", params["key"] )
    assert_match( /#{image}/, params["file"].original_filename )
    assert_match( /image/,    params["file"].content_type      )
    assert_match( /jpeg/,     params["file"].content_type      )
    File.open( "test/web/testfiles/#{image}" ) { |test_image|
      test_image.binmode
      assert_equal( test_image.read, params["file"].read )
    }
    assert_equal( [ [177667, 0],
                   [177667, 10240],
                   [177667, 177667] ],
                  progress_events )
  ensure
    Web::config[:multipart_progress_hook] = nil
  end
  
  ### TODO: What happens if you upload two files?
  ###       What if they have the same name?
  def test_parse_two_uploaded_files
    
  end

  def test_downcase_env
    ENV["KEY"] = "value"
    env = Web::Connection::downcase_env( ENV )
    ["key", "kEy", "kEY", "KEY", "KEy", "Key", "keY"].each{ |k|
      assert( env.has_key?(k) )
      assert_equal( "value", env[k] )
    }
  end

  def options
    {}
  end

  def test_mac_ie_multipart_bug
    require 'yaml'
    test_env = nil
    File.open("test/web/testfiles/mac_ie_env.yaml", "r") { |f|
      test_env = YAML.load(f)
    }

    backup_env = {}
    test_env.each{ |k,v|
      backup_env[k.downcase] = ENV[k.downcase]
      ENV[k.downcase] = v
    }
    
    begin
      File.open("test/web/testfiles/mac_ie_stdin.bin", "r") { |f|
        $stdin = f
        
        cgi.parse_request
        
        assert_equal "test",      cgi.post["text"].first
        assert_equal Web::Upload, cgi.post["small"].first.class
        assert_equal 15,          cgi.post["button.x"].first.to_i
        assert_equal 11,          cgi.post["button.y"].first.to_i
      }
    ensure
      backup_env.each{ |k,v|
        ENV[k] = v
      }
    end
  end
end
