require 'web/wiki'

class HistoryTest < Test::Unit::TestCase
  include Web::Testing

  def do_edit( name, content, comment="" )
    do_request "wiki.rb", "action" => "edit", "page.name" => name
    do_submit "edit_page", "page.content" => content, "page.comment" => comment, "action" => "Save"
  end

  def test_homepage_uses_special_template
    do_request "wiki.rb", "page.name" => "HomePage"
    assert_template_used( "home.html" )
  end

  def test_comments
    do_edit("HomePage", "edit 1", "comment 1")
    Web::local.assert_includes( "page" => { "comment" => "" } )

    do_edit("HomePage", "edit 2", "comment 2")
    Web::local.assert_includes( "page" => { "comment" => "",
                                      "history" => [ { "name" => "HomePage",
				      		       "comment" => "comment 2", },
						     { "name" => "HomePage",
						       "comment" => "comment 1" } ] } )
  end
  
  def test_recent
    10.times{ |i|
      do_edit( "Page#{i}", "edit " + i.to_s, "" )
    }
    Web::local.assert_includes( "wiki" => { "recent" => [{"name" => "Page9", },
                                                         {"name" => "Page8"  },
                                                         {"name" => "Page7"  },
                                                         {"name" => "Page6"  } ] } )

    assert_equal( 10, Web::local['wiki'].recent.size )

    10.times{ |i|
      do_edit( "Page#{i}", "edit " + i.to_s, "" )
    }

    assert_equal( 10,
                  Web::local['wiki'].recent.size,
                  "wiki.recent should be a set of unique pages"  )

  end

  def test_news
    10.times{ |i|
      do_edit( "HomePage", "edit " + i.to_s, "comment " + i.to_s )
    }
    Web::local.assert_includes( "wiki" => { "news" => [{"name" => "HomePage",
                                                  "comment" => "comment 9", },
                                                 {"name" => "HomePage",
                                                  "comment" => "comment 8" },
                                                 {"name" => "HomePage",
                                                  "comment" => "comment 7" },
                                                 {"name" => "HomePage",
                                                  "comment" => "comment 6" } ] } )
  end

  def todo_test_more_news
    do_edit("HomePage", "edit 1", "comment 1")
    do_edit("AnotherPage", "edit 2", "comment 2")

    Web::local.assert_includes( "wiki" => { "more_news" => [ { "name" => "AnotherPage",
    			  	      		       	      "comment" => "comment 2", },
						            { "name" => "HomePage",
						              "comment" => "comment 1" } ] } )
    
    do_request "wiki.rb", "action" => "more_news"
    assert_template_used "more_news.html"

    do_request "wiki.rb", "action" => "more_news_rss"
    assert_template_used "more_news_rss.html"
    assert_tempalte_not_used "template.html"
  end

  def test_more_news_doesnt_have_content_or_history
    do_edit("HomePage", "edit 1", "comment 1")
    do_edit("AnotherPage", "edit 2", "comment 2")

    Web::local.assert_includes( "wiki" => { "more_news" => [ { "name" => "AnotherPage",
    			  	      		       	      "comment" => "comment 2",
							      "content" => "",
							      "history" => [ ] },
						            { "name" => "HomePage",
						              "comment" => "comment 1",
							      "content" => "",
							      "history" => [ ] } ] } )
  end

  def test_revision_number
    name = "RevisionNumber"
    content = "test"
    
    3.times { |i|
      do_edit(name, "edit " + i.to_s)
    }

    do_request "wiki.rb", "page.name" => name
    Web::local.assert_includes( "page" => {"content" => "edit 2",
                                     "revision" => 3 })
    
    do_request "wiki.rb", "action" => "edit", "page.name" => name
    Web::local.assert_includes( "page" => { "revision" => 3 } )
    
    do_submit( "edit_page",
               "page.content" => "final edit",
               "action" => "Save" )
    
    Web::local.assert_includes( "page" => { "revision" => 4 } )
    
  end

  def test_more_news_is_limited_in_size
    assert_equal( 25, Web::Wiki::Page::max_revisions )
    Web::Wiki::Page.module_eval {
      @@max_revisions = 3
    }
    depth = Web::Wiki::Page::max_revisions
    dif = 2
    test  = depth + dif
    
    test.times { |i|
      do_edit("HomePage", "", "edit " + i.to_s)
    }

    expected = (dif...test).to_a.reverse.collect{ |i|
      { "name" => "HomePage",
        "comment" => "edit " + i.to_s,
	"revision" => i + 1 }
    }

    Web::local.assert_includes( "wiki" => {
                                  "more_news" => expected } )

  end

  def test_more_news_doesnt_include_empty_comments
    do_edit("HomePage", "first edits",  "comment 1" )
    do_edit("HomePage", "second edits", ""          )
    do_edit("HomePage", "third edits",  "comment 3" )
  
    Web::local.assert_includes( "wiki" => { "more_news" => [{ "name" => "HomePage",
                                                        "comment" => "comment 3",
                                                        "content" => "",
                                                        "history" => [ ] },
                                                      { "name" => "HomePage",
                                                        "comment" => "comment 1",
                                                        "content" => "",
                                                        "history" => [ ] } ] } )
  end

  def test_history
    do_edit("HomePage", "first edits")
    do_edit("HomePage", "second edits")
    do_edit("HomePage", "third edits")

    do_request( "wiki.rb", "page.name" => "HomePage", "action" => "page_history" )
    assert_template_used( "page_history.html" )
    Web::local.assert_includes( "page" => {
                            "name" => "HomePage",
                            "content" => "third edits",
			    "revision" => 3,
                            "history" => [ { "name" => "HomePage",
                                             "content" => "third edits",
					     "revision" => 3 },
					   { "name" => "HomePage",
                                             "content" => "second edits",
					     "revision" => 2 },
                                           { "name" => "HomePage",
                                             "content" => "first edits",
					     "revision" => 1 },
					    ] } )
  end

  def test_view_history
    do_edit("SomePage", "first edits")
    do_edit("SomePage", "second edits")
    do_edit("SomePage", "third edits")

    do_request( "wiki.rb", "page.name" => "SomePage",
                                 "action" => "view_revision",
				 "revision" => "2" )
    assert_template_used( "basic.html" )
    Web::local.assert_includes( "page" => {
                            "name" => "SomePage",
                            "content" => "second edits",
			    "revision" => 2,
                            "history" => [ { "name" => "SomePage",
                                             "content" => "third edits",
					     "revision" => 3 },
                                           { "name" => "SomePage",
                                             "content" => "second edits",
					     "revision" => 2 },
                                           { "name" => "SomePage",
                                             "content" => "first edits",
					     "revision" => 1 }, ] } )
  end

  def test_edit_history
    do_edit("HomePage", "first edits")
    do_edit("HomePage", "second edits")
    do_edit("HomePage", "third edits")

    do_request( "wiki.rb",
                "page.name" => "HomePage",
                "action" => "edit",
                "revision" => "2" )
    Web::local.assert_includes( "page" =>
                                { "name" => "HomePage",
                                 "content" => "second edits",
                                 "revision" => 2,
                                 "history" => [{"name" => "HomePage",
                                                "content" => "third edits",
                                                "revision" => 3 },
                                               {"name" => "HomePage",
                                                "content" => "second edits",
                                                "revision" => 2 },
                                               {"name" => "HomePage",
                                                "content" => "first edits",
                                                "revision" => 1 }, ] } )
    assert_form_includes( "edit_page",
                          "page" =>
                          {"name" => "HomePage",
                           "content" => "second edits", } )
    do_submit( "edit_page", "page.content" => "fourth edits" )
    
    Web::local.assert_includes( "page" =>
                                {"name" => "HomePage",
                                 "content" => "fourth edits",
                                 "revision" => 4,
                                 "history" => [{"name" => "HomePage",
                                                "content" => "fourth edits",
                                                "revision" => 4 }, 
                                               {"name" => "HomePage",
                                                "content" => "third edits",
                                                "revision" => 3 },
                                               {"name" => "HomePage",
                                                "content" => "second edits",
                                                "revision" => 2 },
                                               {"name" => "HomePage",
                                                "content" => "first edits",
                                                "revision" => 1 }, ] } )
  end

  def test_edit_history_clears_comments
    do_edit("HomePage", "first edits",  "comment 1")
    do_edit("HomePage", "second edits", "comment 2")
    do_edit("HomePage", "third edits",  "comment 3")

    do_request( "wiki.rb", "page.name" => "HomePage",
                                 "action" => "edit",
				 "revision" => "2" )
    Web::local.assert_includes( "page" => {"name" => "HomePage",
                                     "content" => "second edits",
                                     "revision" => 2,
                                     "comment" => "" } )
  end

  def test_history_is_limited_in_size
    assert_equal( 25, Web::Wiki::Page::max_revisions )
    Web::Wiki::Page.module_eval {
      @@max_revisions = 3
    }
    depth = Web::Wiki::Page::max_revisions
    dif = 2
    test  = depth + dif
    
    test.times { |i|
      do_edit("HomePage", "edit " + i.to_s)
    }

    expected = (dif...test).to_a.reverse.collect{ |i|
      { "name" => "HomePage",
        "content" => "edit " + i.to_s,
	"revision" => i + 1 }
    }

    Web::local.assert_includes( "page" => {"name" => "HomePage",
                                           "content" => "edit " + (test-1).to_s,
                                           "revision" => test,
                                           "history" => expected } )
  end

  def setup
    setup_wiki
  end

  def teardown
    teardown_wiki
  end

end

